import os
from sys import exit
from platform import system
from re import match, sub
from datetime import datetime
from loguru import logger
from configparser import ConfigParser

from rsack.exceptions import InvalidURL

class Settings:
    def __init__(self, check=False):
        self.ini_path = os.path.join(get_settings_path(), 'rsack_settings.ini')
        # If settings doesn't exist then create one
        if not os.path.isfile(self.ini_path):
            logger.debug(f'Writing {self.ini_path}')
            self.generate_settings()
        # Read settings.ini
        self.config = ConfigParser()
        self.config.read(self.ini_path)

    @logger.catch
    def generate_settings(self):
        """
        Generates settings.ini based on user input
        """
        # Write file
        config = ConfigParser()
        config['Bugs'] = {'email': "email@protonmail.com",
                          'password': "mypassword",
                          'threads': "2",
                          'path': "C:\Music\Korean",
                          'timed_lyrics': "true",
                          'contributions': "false",
                          "cover_size": "original",
                          "template": "/{artist}/{artist} - {title}"}
        
        config['Genie'] = {'username': "username",
                           'password': "mypassword",
                           'threads': "2",
                           'path': "C:\Music\Korean",
                           'artist_folders': "Y",
                           'timed_lyrics': "Y",
                           'contributions': "N",
                           "template": "/{artist}/{artist} - {title}"}
        
        config['KKBox'] = {"email": "email@protonmail.com",
                           "password": "mypassword",
                           "threads": "2",
                           "path": "C:\Music\KKBox",
                           "template": "/{artist}/{artist} - {title}"}
        
        with open(self.ini_path, 'w+') as configfile:
            config.write(configfile)
        exit()

    def _str_to_boolean(self, section: str, d: dict) -> dict:
        """Iterates dict converting possible boolean values

        Args:
            section (str): Section header
            d (dict): Dictionary to iterate

        Returns:
            dict: Dictionary with string values replaced with boolean
        """
        for k in d:
            try:
                d[k] = self.config.getboolean(section, k)
            except ValueError: # Not a boolean
                pass
        return d
        
    def Bugs(self) -> dict:
        """
        Returns the contents of the 'Bugs' section of settings.ini as dict.
        """
        return self._str_to_boolean(section='Bugs', d=dict(self.config.items('Bugs')))

    def Genie(self) -> dict:
        """
        Returns the contents of the 'Genie' section of settings.ini as dict.
        """
        return self._str_to_boolean(section='Genie', d=dict(self.config.items('Genie')))
    
    def KKBox(self) -> dict:
        """
        Returns the contents of the 'KKBox' section of settings.ini as dict.
        """
        return self._str_to_boolean(section='KKBox', d=dict(self.config.items('KKBox')))


def track_to_flac(track: dict, album: dict, lyrics: str) -> dict:
    """Creates dict with appropriate FLAC headers.

    Args:
        track (dict): Dict containing track info from API response
        album (dict): Dict containing album info from API response
        lyrics (str): Lyrics of the track
    """
    meta = {
        "ALBUM": track['album_title'],
        "ALBUMARTIST": album['artist_disp_nm'],
        "ARTIST": track['artist_disp_nm'],
        "TITLE": track['track_title'],
        "DISCNUMBER": str(track['disc_id']),
        "DISCTOTAL": str(album['disc_total']),
        "TRACKNUMBER": str(track['track_no']),
        "TRACKTOTAL": str(track['track_total']),
        "COMMENT": str(track['track_id']),
        "DATE": _format_date(track['release_ymd']),
        "GENRE": album['genre_str'],
        "LABEL": '; '.join(str(label['label_nm']) for label in album['labels']),
        "LYRICS": lyrics
    }
    return meta

def _format_date(date):
    """Formats album release date to preferred format"""
    # Append release day if not present.
    if len(date) == 6:
        date = date + "01"
    date_patterns = ["%Y%m%d", "%Y%m", "%Y"]
    for pattern in date_patterns:
        try:
            return datetime.strptime(date, pattern).strftime('%Y.%m.%d')
        except ValueError:
            pass

def bugs_id(url: str) -> str:
    return match(
        r'https?://music\.bugs\.co\.kr/(?:(?:album|artist|track|playlist)/|[a-z]{2}-[a-z]{2}-?\w+(?:-\w+)*-?)(\w+)',url).group(1)

def genie_id(url: str) -> match:
    expression = r"https://genie.co.kr/detail/(artistInfo|albumInfo)......([0-9]*)"
    result = match(expression, url)
    # This shouldn't be needed, regex needs to be fixed.
    if not result:
        expression = r"https://www.genie.co.kr/detail/(artistInfo|albumInfo)......([0-9]*)"
        result = match(expression, url)
    if result:
        return result
    raise InvalidURL(url)
 
def get_settings_path() -> str:
    """Returns path of home folder to store settings.ini"""
    if "XDG_CONFIG_HOME" in os.environ:
        return os.environ['XDG_CONFIG_HOME']
    elif "HOME" in os.environ:
        return os.environ['HOME']
    elif "HOMEDRIVE" in os.environ and "HOMEPATH" in os.environ:
        return os.environ['HOMEDRIVE'] + os.environ['HOMEPATH']
    else:
        return os.path._getfullpathname("./")

def get_ext(type: str) -> str:
    """Return the filetype.

    Args:
        type (str): "FILE_EXT" from player/j_StmInfo.json response
            Known FILE_EXT's:
                        F96: FLAC 24bit/96kHz
                        F44: FLAC 24bit/44.1kHz
                        FLA: FLAC 16bit
                        MP3: MP3
                        
    Returns:
        str: File extension
    """
    if type == "MP3":
        return ".mp3"
    else:
        return ".flac"

def insert_total_tracks(tracks: list[dict]):
    """Add total_tracks to track metadata"""
    total_tracks_by_disc_id = {}
    for track in tracks:
        if track["disc_id"] not in total_tracks_by_disc_id:
            total_tracks_by_disc_id[track["disc_id"]] = 0
        total_tracks_by_disc_id[track["disc_id"]] += 1

    for track in tracks:
        track["track_total"] = total_tracks_by_disc_id[track["disc_id"]]

def _is_win() -> bool:
	if system() == 'Windows':
		return True

def sanitize(fn: str) -> str:
    """Sanitizes filenames based on Operating System"""
    if _is_win():
        return sub(r'[\/:*?"><|]', '_', fn).strip('. ')
    else:
        return sub('/', '_', fn).strip()

def contribution_check(artist_id_provided: int, artist_id_api: int) -> bool:
    """Checks if artist is contributing"""
    if artist_id_provided == artist_id_api:
        return False
    else:
        return True

def format_genie_lyrics(lyrics: dict) -> str:
    """Convert Genie dict to a usable str format for tagging"""
    # Convert millisecond keys to format [00:00.00][minutes:seconds.milliseconds]
    lines = [f"[{datetime.fromtimestamp(int(x)/1000).strftime('%M:%S.%f')[:-4]}]{lyrics[x]}" for x in lyrics]
    return '\n'.join(lines)


import os
import mutagen.id3 as id3
from loguru import logger
from urllib.parse import unquote
from mutagen.flac import FLAC, Picture
from concurrent.futures import ThreadPoolExecutor
from requests.models import Response

from rsack.clients import genie
from rsack.utils import Settings, contribution_check, format_genie_lyrics, get_ext, sanitize, _format_date


class Download:
    def __init__(self, type: str, id: int):
        """Initialize and control the flow of the download"""
        self.settings = Settings().Genie()
        self.client = genie.Client()
        self.client.auth(username=self.settings['username'], password=self.settings['password'])
        logger.info(f"Threads: {self.settings['threads']}")
        if type == "artist":
            self._artist(id)
        elif type == "album":
            self._album(id)

    def _artist(self, id: int):
        """Iterate albums in artist"""
        meta = self.client.get_artist_albums(id)
        for album in meta['DataSet']['DATA']:
            if contribution_check(id, int(album['ARTIST_ID'])):
                if self.settings['contributions'] == "Y":
                    self._album(album['ALBUM_ID'])
                else:
                    logger.debug("Skipping contribution")
            else:
                self._album(album['ALBUM_ID'])

    @logger.catch
    def _template(self):
        keys = {
            "artist": unquote(self.album['ARTIST_NAME']),
            "title": unquote(self.album['ALBUM_NAME']),
            "date": self.album['ALBUM_RELEASE_DT'],
            "album_id": str(self.album['ALBUM_ID']),
            "artist_id": str(self.album['ARTIST_ID']),
            "type": self.album['ALBUM_TYPE'],
            "label_name": self.album['ALBUM_PLANNER'],
        }
        template = self.settings['template']
        for k in keys:
            template = template.replace(f"{{{k}}}", sanitize(keys[k]))
        return template
    
    @logger.catch
    def _album(self, id: int):
        """Iterate tracks in album"""
        self.meta = self.client.get_album(id)
        self.album = self.meta['DATA0']['DATA'][0]
        self._template()
        logger.info(f"Album: {unquote(self.album['ALBUM_NAME'])}")
        self.album_path = self.settings['path'] + self._template()
        
        try:
            if not os.path.isdir(self.album_path):
                logger.debug(f"Creating: {self.album_path}")
                os.makedirs(self.album_path)
        except OSError as exc:
            if exc.errno == 36: # Exceeded path limit
                self.album_path = os.path.join(self.settings['path'], "EDIT ME")
                if not os.path.isdir(self.album_path): # Retry
                    logger.debug(f"Creating: {self.album_path}")
                    os.makedirs(self.album_path)
                    
        # Create disc directories
        self.disc_total = int(self.meta['DATA1']['DATA'][len(self.meta['DATA1']['DATA']) - 1]['ALBUM_CD_NO'])
        if  self.disc_total > 1:
            for i in range(0, self.disc_total):
                d = os.path.join(self.album_path, f"Disc {i + 1}")
                if not os.path.isdir(d):
                    os.makedirs(d)
                    
        cover_url = unquote(self.album['ALBUM_IMG_PATH_600'])
        if cover_url == "":
            cover_url = unquote(self.album['ALBUM_IMG_PATH'])
        self._download_cover(cover_url)

        # Initialize empty lists
        track_ids = []
        track_numbers = []
        disc_numbers = []
        track_artist = []
        # Append required information to their relevant lists
        for track in self.meta['DATA1']['DATA']:
            track_ids.append(int(track['SONG_ID']))
            track_numbers.append(f"{track['ALBUM_TRACK_NO']}")
            disc_numbers.append(track['ALBUM_CD_NO'])
            track_artist.append(track['ARTIST_NAME'])

        # Create ThreadPoolExecutor to handle multiple downloads at once
        with ThreadPoolExecutor(max_workers=int(self.settings['threads'])) as executor:
            executor.map(self._track, track_ids, track_numbers,
                         disc_numbers, track_artist)

    @logger.catch
    def _track(self, id: int, track_number: str, disc_number: str, track_artist: str):
        """Handles the download of a track

        Args:
            id (int): Unique ID of the track
            track_number (str): String representation of the track number
            disc_number (str): String representation of the disc number
            track_artist (str): Name of the artist connected to the track
        """
        meta = self.client.get_stream_meta(id)
        if meta: # Meta can return False if unavailable for stream
            logger.info(f"Track: {unquote(meta['SONG_NAME'])}")
            ext = get_ext(meta['FILE_EXT'])
            if self.disc_total > 1:
                file_path = os.path.join(self.album_path, f"Disc {disc_number}", f"{int(track_number):02d}. {sanitize(unquote(meta['SONG_NAME']))}{ext}")
            else:
                file_path = os.path.join(self.album_path, f"{int(track_number):02d}. {sanitize(unquote(meta['SONG_NAME']))}{ext}")
            if os.path.exists(file_path):
                logger.debug(f"{file_path} already exists.")
            else:
                r = self.client.session.get(unquote(meta['STREAMING_MP3_URL']))
                r.raise_for_status()
                lyrics = self.client.get_timed_lyrics(id)
                try:
                    self._write_track(file_path, r)
                except OSError:
                    # OSError assumes excessive file length, rename file and continue writing
                    file_path = os.path.join(self.album_path, f"{track_number}.{ext}")
                    logger.debug(f"{track_number} has been renamed as it exceeded the maximum length.")
                    self._write_track(file_path, r)
                self._fix_tags(file_path, lyrics, ext, track_number, disc_number,
                            track_artist, unquote(meta['SONG_NAME']))

    @staticmethod
    def _write_track(file_path: str, r: Response):
        """Write track response data to file"""
        with open(file_path, 'wb') as f:
            for chunk in r.iter_content(32 * 1024):
                if chunk:
                    f.write(chunk)

    @logger.catch
    def _fix_tags(self, path: str, lyrics: str, ext: str, track_number: str, disc_number: str, track_artist: str, track_title: str):
        """Fixes I3D/FLAC metadata

        Args:
            path (str): Path of .mp3/.flac file
            lyrics (str): Song lyrics
            ext (str): File extension (.mp3/.flac)
            track_number (str): String representation of the track number
            disc_number (str): String representation of the disc number
            track_artist (str): Name of the artist connected to the track
            track_title (str): Name of the track
        """
        if ext == ".mp3":
            # Instantiate ID3 object
            try:
                audio = id3.ID3(path)
            except id3.ID3NoHeaderError:
                audio = id3.ID3()
            # Delete pre-embedded artwork
            audio.delall("APIC")
            # Embed existing artwork
            if self.cover_path:
                with open(self.cover_path, 'rb') as cov_obj:
                    audio.add(id3.APIC(3, 'image/jpg', 3, '', cov_obj.read()))
            # Append necessary tags
            audio['TIT2'] = id3.TIT2(text=track_title)
            audio['TALB'] = id3.TALB(text=unquote(self.album['ALBUM_NAME']))
            audio['TCON'] = id3.TCON(text=unquote(self.album['ALBUM_NAME']))
            audio['TRCK'] = id3.TRCK(text=str(track_number) + "/" + str(len(self.meta['DATA1']['DATA'])))
            audio['TPOS'] = id3.TPOS(text=str(disc_number) + "/" + str(self.disc_total))
            audio['TDRC'] = id3.TDRC(text=self.album['ALBUM_RELEASE_DT'])
            audio['TPUB'] = id3.TPUB(text=unquote(self.album['ALBUM_PLANNER']))
            audio['TPE1'] = id3.TPE1(text=unquote(track_artist))
            audio['TPE2'] = id3.TPE2(text=unquote(self.album['ARTIST_NAME']))
            audio['TCON'] = id3.TCON(text="")
            audio.setall("COMM", [id3.COMM(text=[u"지니뮤직"], encoding=id3.Encoding.UTF8)])
            if lyrics != None and self.settings['timed_lyrics']:
                    lyrics = [(v, int(k)) for k, v in lyrics.items()]
                    audio.setall("SYLT", [id3.SYLT(encoding=id3.Encoding.UTF8, lang='eng', format=2, type=1, text=lyrics)])
            logger.debug(f"Writing tags to: {path}")
            audio.save(path, "v2_version=3") # Write file
        else:
            audio = FLAC(path)
            # Delete pre-embedded artwork
            audio.clear_pictures()
            # Embed existing artwork
            if self.cover_path:
                f_image = Picture()
                f_image.type = 3
                f_image.desc = 'Front Cover'
                with open(self.cover_path, 'rb') as f:
                    f_image.data = f.read()
                audio.add_picture(f_image)
            # Append necessary tags
            audio['TRACKNUMBER'] = str(track_number)
            audio['TRACKTOTAL'] = str(len(self.meta['DATA1']['DATA']))
            audio['DISCNUMBER'] = str(disc_number)
            audio['DISCTOTAL'] = str(self.disc_total)
            audio['DATE'] = self.album['ALBUM_RELEASE_DT']
            audio['LABEL'] = self.album['ALBUM_PLANNER']
            audio['ARTIST'] = unquote(track_artist)
            audio['ALBUMARTIST'] = unquote(self.album['ARTIST_NAME'])
            audio['ALBUM'] = unquote(self.album['ALBUM_NAME'])
            audio['TITLE'] = track_title
            audio['COMMENT'] = "지니뮤직"
            if lyrics != None and self.settings['timed_lyrics']:
                audio['LYRICS'] = format_genie_lyrics(lyrics)
            logger.debug(f"Writing tags to: {path}")
            audio.save()  # Write file

    def _download_cover(self, url: str):
        """Download cover artwork"""
        self.cover_path = os.path.join(self.album_path, 'cover.jpg')
        if not os.path.isfile(self.cover_path):
            r = self.client.session.get(unquote(url))
            with open(self.cover_path, 'wb') as f:
                f.write(r.content)
        else:
            logger.debug(f"{self.cover_path} already exists.")


import json
import requests

from loguru import logger
from rsack.exceptions import DeviceIDError

class Client:
    def __init__(self):
        self.session = requests.Session()
        self.dev_id = "eb9d53a3c424f961"

        self.session.headers.update({
            "User-Agent": "genie/ANDROID/5.1.1/WIFI/SM-G930L/dreamqltecaneb9d53a3c424f961/500200714/40807",
            "Referer": "app.genie.co.kr"
        })

        self.session.mount('https://', requests.adapters.HTTPAdapter(max_retries=3))
        
    def make_call(self, sub: str, epoint: str, data: dict) -> dict:
        """Makes API call to specified endpoint

        Args:
            sub (str): Subdomain
            epoint (str): Endpoint
            data (dict): POST data
        
        Endpoints used:
            player/j_StmInfo.json: Returns track information
            member/j_Member_Login.json: Authentication.
            song/j_AlbumSongList.json: Returns album information

        Returns:
            dict: JSON Response
        """
        try:
            r = self.session.post("https://{}.genie.co.kr/{}".format(sub, epoint), data=data)
        except requests.exceptions.ConnectionError:
            logger.debug("Remote end closed connection, retrying.")
            r = self.session.post("https://{}.genie.co.kr/{}".format(sub, epoint), data=data)
        return r.json()

    def auth(self, username: str, password: str):
        """
        Authenticate session
        """
        data = {
            "uxd": username,
            "uxx": password
        }
        r = self.make_call("app", "member/j_Member_Login.json", data)
        if r['Result']['RetCode'] != "0":
            logger.critical(f"Authentication Failed: {r['RetMsg']}")
            exit()
        else:
            logger.info("Login Successful.")
        self.usr_num = r['DATA0']['MemUno']
        self.usr_token = r['DATA0']['MemToken']
        self.stm_token = r['DATA0']['STM_TOKEN']

    def get_album(self, id: int) -> dict:
        """Retrieve album information"""
        data = {
            "axnm": id,
            "dcd": self.dev_id,
            "mts": "Y",
            "stk": self.stm_token,
            "svc": "IV",
            "tct": "Android",
            "unm": self.usr_num,
            "uxtk": self.usr_token
        }
        r = self.make_call("app", "song/j_AlbumSongList.json", data)
        if r['Result']['RetCode'] != "0":
            logger.critical("Failed to retrieve metadata")
        return r
    
    def get_artist_albums(self, id: int) -> dict:
        """Retrieve artists album information"""
        data = {
            "uxtk": self.usr_token,
            "sign": "Y",
            "tct": "Android",
            "svc": "IV",
            "stk": self.stm_token,
            "dcd": self.dev_id,
            "xxnm": id,
            "unm": self.usr_num,
            "mts": "Y",
            "pgsize": 500
        }
        r = self.make_call("app", "song/j_ArtistAlbumList.json", data)
        if r['Result']['RetCode'] != "0":
            logger.critical("Failed to retrieve metadata")
        return r

    def get_artist(self, id: int) -> dict:
        """Retrieves artist information"""
        data = {
            "uxtk": self.usr_token,
            "sign": "Y",
            "tct": "Android",
            "svc": "IV",
            "stk": self.stm_token,
            "dcd": self.dev_id,
            "xxnm": id,
            "unm": self.usr_num,
            "mts": "Y"
        }
        r = self.make_call("info", "info/artist", data)
        if r['result']['ret_code'] != "0":
            logger.critical("Failed to retrieve metadata")
        return r
        
    def get_stream_meta(self, id: int) -> dict:
        """Retrieves information on a streamable track

        Args:
            id (int): Unique ID of track

        Raises:
            DeviceIDError: Raises when RetCode "A00003" is returned.
                        Caused by sudden change in DeviceID.
                              
        Returns:
            dict: JSON Response
        """
        data = {
            "bitrate": "24bit",
            "sign": "Y",
            "mts": "Y",
            "dcd": self.dev_id,
            "stk": self.stm_token,
            "itn": "Y",
            "svc": "IV",
            "unm": self.usr_num,
            "uxtk": self.usr_token,
            "xgnm": id,
            "apvn": 40807
        }
        r = self.make_call("stm", "player/j_StmInfo.json", data)
        if r['Result']['RetCode'] == "A00003":
            raise DeviceIDError("Device ID has been changed since last stream.")
        if r['Result']['RetCode'] != "0":
            logger.critical("Failed to retrieve metadata")
        if r['Result']['RetCode'] == "S00001":
            logger.debug("This content is currently unavailable for service")
            return False
        return r['DataSet']['DATA'][0]
    
    def get_timed_lyrics(self, id: str) -> dict:
        """Retrieve the timed lyrics for a track"""
        r = self.session.get(f"https://dn.genie.co.kr/app/purchase/get_msl.asp?songid={id}&callback=GenieCallBack")
        if r.content.decode('utf-8') == 'NOT FOUND LYRICS':
            return None
        # Remove unwanted characters
        r = r.content.decode('utf-8')[14:-2]
        return json.loads(r)


import json
import requests
from loguru import logger
from time import time, sleep
from random import randrange
from Cryptodome.Cipher import ARC4
from Cryptodome.Hash import MD5
from bs4 import BeautifulSoup

class KkboxAPI:
    """KKBox Client
    Source: https://github.com/uhwot/orpheusdl-kkbox/blob/1e217c41af599a8ff9a821d8e3ac13535a03c6de/kkapi.py
    """
    def __init__(self, kc1_key="7f1a68f00b747f4ac1469c72e7ef492c", kkid=None):
        self.kc1_key = kc1_key.encode('ascii')
        self.s = requests.Session()
        self.s.headers.update({
            'user-agent': 'okhttp/3.14.9'
        })

        self.kkid = kkid or '%032X' % randrange(16**32)

        self.params = {
            'enc': 'u',
            'ver': '06090076',
            'os': 'android',
            'osver': '11',
            'lang': 'en',
            'ui_lang': 'en',
            'dist': '0021',
            'dist2': '0021',
            'resolution': '411x683',
            'of': 'j',
            'oenc': 'kc1',
        }

    def kc1_decrypt(self, data):
        cipher = ARC4.new(self.kc1_key)
        return cipher.decrypt(data).decode('utf-8')

    def get_date(self, id):
        logger.debug("Scraping release date")
        r = self.s.get(f"https://www.kkbox.com/hk/tc/album/{id}")
        soup = BeautifulSoup(r.content, 'html.parser')
        return soup.find('div', attrs={'class': "date"}).text.replace("/", ".")
        
    def api_call(self, host, path, params={}, payload=None):
        if host == 'ticket':
            payload = json.dumps(payload)

        params.update(self.params)
        params.update({'timestamp': int(time())})

        url = f'https://api-{host}.kkbox.com.tw/{path}'
        if not payload:
            r = self.s.get(url, params=params)
        else:
            r = self.s.post(url, params=params, data=payload)

        resp = json.loads(self.kc1_decrypt(r.content)) if r.content else None
        return resp

    def login(self, email, password, region_bypass=False):
        md5 = MD5.new()
        md5.update(password.encode('utf-8'))
        pswd = md5.hexdigest()

        host = 'login' if not region_bypass else 'login-utapass'

        resp = self.api_call(host, 'login.php', payload={
            'uid': email,
            'passwd': pswd,
            'kkid': self.kkid,
            'registration_id': '',
        })

        if not resp and region_bypass:
            logger.critical('Account expired')

        if resp['status'] not in (2, 3, -4):
            logger.critical("Credentials invalid")

        if resp['status'] == -4 and not region_bypass: # Region locked
            return self.login(email, password, region_bypass=True)

        self.region_bypass = region_bypass

        self.apply_session(resp)
        logger.info("Logged in")

    def renew_session(self):
        host = 'login' if not self.region_bypass else 'login-utapass'
        resp = self.api_call(host, 'check.php')
        if resp['status'] not in (2, 3, -4):
            logger.critical('Session renewal failed')
        self.apply_session(resp)

    def apply_session(self, resp):
        self.sid = resp['sid']
        self.params['sid'] = self.sid

        self.lic_content_key = resp['lic_content_key'].encode('ascii')

        self.available_qualities = ['128k', '192k', '320k']
        if resp['high_quality']:
            self.available_qualities.append('hifi')
            self.available_qualities.append('hires')

    def get_songs(self, ids):
        resp = self.api_call('ds', 'v2/song', payload={
            'ids': ','.join(ids),
            'fields': 'artist_role,song_idx,album_photo_info,song_is_explicit,song_more_url,album_more_url,artist_more_url,genre_name,is_lyrics,audio_quality'
        })
        if resp['status']['type'] != 'OK':
            logger.critical('Track not found')
        return resp['data']['songs']

    def get_song_lyrics(self, id):
        logger.debug(f"Retrieving lyrics: {id}")
        return self.api_call('ds', f'v1/song/{id}/lyrics')

    def get_album(self, id):
        resp = self.api_call('ds', f'v1/album/{id}')
        if resp['status']['type'] != 'OK':
            logger.critical('Album not found')
        return resp['data']

    def get_album_more(self, raw_id):
        return self.api_call('ds', 'album_more.php', params={
            'album': raw_id
        })

    def get_artist(self, id):
        resp = self.api_call('ds', f'v3/artist/{id}')
        if resp['status']['type'] != 'OK':
            logger.critical('Artist not found')
        return resp['data']
    
    def get_artist_albums(self, raw_id, limit, offset):
        resp = self.api_call('ds', f'v2/artist/{raw_id}/album', params={
            'limit': limit,
            'offset': offset,
        })
        if resp['status']['type'] != 'OK':
            logger.critical('Artist not found')
        return resp['data']['album']

    def get_playlists(self, ids):
        resp = self.api_call('ds', f'v1/playlists', params={
            'playlist_ids': ','.join(ids)
        })
        if resp['status']['type'] != 'OK':
            logger.critical('Playlist not found')
        return resp['data']['playlists']

    def search(self, query, types, limit):
        return self.api_call('ds', 'search_music.php', params={
            'sf': ','.join(types),
            'limit': limit,
            'query': query,
            'search_ranking': 'sc-A',
        })

    def get_ticket(self, song_id, play_mode = None):
        resp = self.api_call('ticket', 'v1/ticket', payload={
            'sid': self.sid,
            'song_id': song_id,
            'ver': '06090076',
            'os': 'android',
            'osver': '11',
            'kkid': self.kkid,
            'dist': '0021',
            'dist2': '0021',
            'timestamp': int(time()),
            'play_mode': play_mode,
        })

        if resp['status'] != 1:
            if resp['status'] == -1:
                self.renew_session()
                return self.get_ticket(song_id, play_mode)
            elif resp['status'] == -4:
                self.auth_device()
                return self.get_ticket(song_id, play_mode)
            elif resp['status'] == 2:
                # tbh i'm not sure if this is some rate-limiting thing
                # or if it's a bug on their slow-as-hell servers
                sleep(0.5)
                return self.get_ticket(song_id, play_mode)
            logger.critical("Couldn't get track URLs")

        return resp['uris']

    def auth_device(self):
        resp = self.api_call('ds', 'active_sid.php', payload={
            'ui_lang': 'en',
            'of': 'j',
            'os': 'android',
            'enc': 'u',
            'sid': self.sid,
            'ver': '06090076',
            'kkid': self.kkid,
            'lang': 'en',
            'oenc': 'kc1',
            'osver': '11',
        })
        if resp['status'] != 1:
            logger.critical("Couldn't auth device")

    def kkdrm_dl(self, url, path):
        # skip first 1024 bytes of track file
        resp = self.s.get(url, stream=True, headers={'range': 'bytes=1024-'})
        resp.raise_for_status()

        size = int(resp.headers['content-length'])

        # drop 512 bytes of keystream
        rc4 = ARC4.new(self.lic_content_key, drop=512)

        with open(path, 'wb') as f:
            for chunk in resp.iter_content(chunk_size=4096):
                f.write(rc4.decrypt(chunk))



import os
import requests
from datetime import datetime
from loguru import logger
from mutagen.flac import FLAC, Picture
import mutagen.id3 as id3
from mutagen.id3 import ID3NoHeaderError
from concurrent.futures import ThreadPoolExecutor

from rsack.clients import bugs
from rsack.utils import Settings, track_to_flac, insert_total_tracks, contribution_check, sanitize, _format_date


class Download:
    def __init__(self, type: str, id: int):
        """Initialize and control flow of download process

        Args:
            type (str): String to declare type of download (album/artist/track).
            id (int): Unique ID.
        """
        self.settings = Settings().Bugs()
        self.client = bugs.Client()
        self.conn_info = self.client.auth(email=self.settings['email'], password=self.settings['password'])
        logger.info(f"Threads: {self.settings['threads']}")
        if type == "artist":
            self._artist(id)
        elif type == "album":
            self._album(id)

    def _artist(self, id: int):
        """Handle artist downloads

        Args:
            id (int): Unique Artist ID
        
        Note:
            self.client.get_artist() returns an album list, but does not include all the necessary tagging info.
            This means it's not worth passing it on and an additional request has to be used.
        """
        artist = self.client.get_artist(id)
        logger.info(f"{len(artist['list'][1]['artist_album']['list'])} releases found")
        for album in artist['list'][1]['artist_album']['list']:
            contribution = contribution_check(id, int(album['artist_id']))
            if contribution:
                if self.settings['contributions']:
                    self._album(album['album_id'])
                else:
                    logger.debug("Skipping album contribution")
            else:
                self._album(album['album_id'])
                
    def _album(self, id: int):
        """Handle album downloads

        Args:
            id (int): Unique album id
        """
        self.album = self.client.get_album(id)['list'][0]['album_info']['result']
        logger.info(f"Album: {self.album['title']}")
        # Acquire disc total
        self.album['disc_total'] = self.album['tracks'][-1]['disc_id']
        # Add track_total to meta.
        insert_total_tracks(self.album['tracks'])
        self._album_path()
        self._download_cover()
        # Begin downloading tracks
        with ThreadPoolExecutor(max_workers=int(self.settings['threads'])) as executor:
            executor.map(self._download, self.album['tracks'])
    
    @logger.catch
    def _template(self):
        keys = {
            "artist": self.album['artist_disp_nm'],
            "title": self.album['title'],
            "local_title": self.album['title_local'],
            "date": _format_date(self.album['release_ymd']),
            "local_date": _format_date(self.album['release_local_ymd']),
            "album_id": str(self.album['album_id']),
            "artist_id": str(self.album['artist_id']),
            "type": self.album['album_tp'],
            "agency_name": self.album['agency_nm'],
            "agency_id": str(self.album['agency_id']),
            "label_name": self.album['labels'][0]['label_nm'],
            "label_id": str(self.album['labels'][0]['label_id'])
        }
        template = self.settings['template']
        for k in keys:
            template = template.replace(f"{{{k}}}", sanitize(keys[k]))
        return template
            
    @logger.catch
    def _album_path(self):
        """Creates necessary directories"""
        self.album_path = self.settings['path'] + self._template()
        try:
            if not os.path.exists(self.album_path):
                logger.debug(f"Creating {self.album_path}")
                os.makedirs(self.album_path)
        except OSError as exc:
            if exc.errno == 36: # Exceeded path limit
                if len(self.album['artist_disp_nm']) > len(self.album['title']): #  Check whether artist name or album name is the issue
                    self.album['artist_disp_nm'] = "Various Artists" # Change to V.A. as Bugs has likely compiled a huge list of artists
                    logger.debug("Artist name forcibly changed to try and reduce length")
                else: # If title is the issue
                    logger.debug("Album title forcibly changed to try and reduce length")
                    self.album_path = self.settings['path'] + "EDIT ME"
                # Retry
                if not os.path.exists(self.album_path):
                    logger.debug(f"Creating {self.album_path}")
                    os.makedirs(self.album_path)
                    
        # Create nested disc folders
        if self.album['disc_total'] > 1:
            self.discs = True
            for i in range(self.album['disc_total']):
                d = os.path.join(self.album_path, f"Disc {str(i + 1)}")
                if not os.path.exists(d):
                    os.makedirs(d)
        else: 
            self.discs = False

    @logger.catch
    def _download(self, track: dict):
        """Downloads track

        Args:
            track (dict): Contains track information from API response
        """
        if track['is_flac_str_premium'] and not self.client.premium:
            logger.warning("Lossless is only available for Premium users, MP3 will be downloaded.")
        logger.info(f"Track: {track['track_title']}")
        if self.discs:
            file_path = os.path.join(self.album_path, f"Disc {str(track['disc_id'])}", f"{track['track_no']:02d}. {sanitize(track['track_title'])}.temp")
        else:
            file_path = os.path.join(self.album_path, f"{track['track_no']:02d}. {sanitize(track['track_title'])}.temp")
        if self._exist_check(file_path):
            logger.debug(f"{track['track_title']} already exists")
        else:
            # Create params required to request the track
            params = {
                "ConnectionInfo": self.conn_info,
                "api_key": self.client.api_key,
                "overwrite_session": "Y",
                "track_id": track['track_id']
            }
            # Create headers for byte position.
            headers = {
                "Range": 'bytes=%d-' % self._return_bytes(file_path),
            }
            r = requests.get(f"http://api.bugs.co.kr/3/tracks/{track['track_id']}/listen/android/flac", headers=headers, params=params, stream=True)
            if r.url.split("?")[0].endswith(".mp3"): # If response redirects to MP3 file set quality to .mp3
                quality = '.mp3'
            elif r.url.split("?")[0].endswith(".m4a"):
                quality = '.m4a' 
            else: # Otherwise .flac
                quality = '.flac'
            if quality != '.m4a':
                if r.status_code == 404:
                    logger.info(f"{track['track_title']} unavailable")
                else:
                    with open(file_path, 'ab') as f:
                        for chunk in r.iter_content(32 * 1024):
                            if chunk:
                                f.write(chunk)
                    c_path = file_path.replace(".temp", quality)
                    os.rename(file_path, c_path)
                    self._tag(track, c_path)
            else:
                logger.info(f"{track['track_title']} is unavailable.")

    @staticmethod
    def _return_bytes(file_path: str) -> int:
        """Returns number of bytes in file

        Args:
            file_path (str): File path

        Returns:
            int: Returns size in bytes
        """
        
        if os.path.exists(file_path):
            logger.debug(f"Existing .temp file {os.path.basename(file_path)} has resumed.")
            return os.path.getsize(file_path)
        else:
            return 0
    
    @staticmethod
    def _exist_check(file_path: str) -> bool:
        """Check if file exists for both possible cases

        Args:
            file_path (str): .temp file path

        Returns:
            bool: True if exists else false
        """
        if os.path.exists(file_path.replace('.temp', '.mp3')):
            return True
        if os.path.exists(file_path.replace('.temp', '.flac')):
            return True
        else:
            return False
    
    def _download_cover(self):
        """Downloads cover artwork"""
        self.cover_path = os.path.join(self.album_path, 'cover.jpg')
        if os.path.exists(self.cover_path):
            logger.info('Cover already exists')
        else:
            r = requests.get(self.album['img_urls'][self.settings['cover_size']])
            r.raise_for_status
            with open(self.cover_path, 'wb') as f:
                f.write(r.content)
            logger.info('Cover artwork downloaded.')
    
    @logger.catch
    def _tag(self, track: dict, file_path: str):
        """Append ID3/FLAC tags

        Args:
            track (dict): API response containing track information
            file_path (str): File being tagged
        """
        lyrics = self._get_lyrics(track['track_id'], track['lyrics_tp'])
        tags = track_to_flac(track, self.album, lyrics)
        if str(file_path).endswith('.flac'):
            f_file = FLAC(file_path)
            # Add cover artwork to flac file
            f_file.clear_pictures() # Delete existing cover artwork
            if self.cover_path:
                f_image = Picture()
                f_image.type = 3
                f_image.desc = 'Front Cover'
                with open(self.cover_path, 'rb') as f:
                    f_image.data = f.read()
                f_file.add_picture(f_image)
            logger.debug(f"Writing tags to {file_path}")
            for k, v in tags.items():
                f_file[k] = str(v)
            f_file.save()
        if str(file_path).endswith('.mp3'):
            # Legend contains all ID3 tags for each FLAC header.
            legend = {
                "ALBUM": id3.TALB,
                "ALBUMARTIST": id3.TPE2,
                "ARTIST": id3.TPE1,
                "COMMENT": id3.COMM,
                "COMPOSER": id3.TCOM,
                "COPYRIGHT": id3.TCOP,
                "DATE": id3.TDRC,
                "GENRE": id3.TCON,
                "ISRC": id3.TSRC,
                "LABEL": id3.TPUB,
                "PERFORMER": id3.TOPE,
                "TITLE": id3.TIT2,
                "LYRICS": id3.USLT
            }
            try:
                m_file = id3.ID3(file_path)
            except ID3NoHeaderError:
                m_file = id3.ID3()
            logger.debug(f"Writing tags to {file_path}")
            # Apply tags using the legend
            for k, v in tags.items():
                try:
                    id3tag = legend[k]
                    m_file[id3tag.__name__] = id3tag(encoding=3, text=v)
                except KeyError:
                    continue
            # Track and disc numbers
            m_file.add(id3.TRCK(encoding=3, text=f"{track['track_no']}/{track['track_total']}"))
            m_file.add(id3.TPOS(encoding=3, text=f"{track['disc_id']}/{self.album['disc_total']}"))
            # Apply cover artwork
            m_file.delall("APIC") # Delete existing cover artwork
            if self.cover_path:
                with open(self.cover_path, 'rb') as cov_obj:
                    m_file.add(id3.APIC(3, 'image/jpg', 3, '', cov_obj.read()))
            m_file.save(file_path, 'v2_version=3')

    def _get_lyrics(self, track_id: int, lyrics_tp: str) -> str:
        """Retrieves and formats track lyrics

        Args:
            track_id (int): Unique track ID.
            lyrics_tp (str): 'T'/'N': Timed/Normal lyrics from settings.

        Returns:
            str: Formatted lyrics
        """
        # If user prefers timed then retrieve timed lyrics
        if lyrics_tp and self.settings['timed_lyrics']:
            # Retrieve timed lyrics
            r = requests.get(f"https://music.bugs.co.kr/player/lyrics/T/{track_id}")
            # Format timed lyrics
            lyrics = r.json()['lyrics'].replace("＃", "\n")
            line_split = (line.split('|') for line in lyrics.splitlines())
            lyrics = ("\n".join(
                f'[{datetime.fromtimestamp(round(float(a), 2)).strftime("%M:%S.%f")[0:-4]}]{b}' for a, b in line_split))
        # If user prefers untimed or timed unavailable then use untimed
        elif not lyrics_tp or not self.settings['timed_lyrics']:
            r = requests.get(f'https://music.bugs.co.kr/player/lyrics/N/{track_id}')
            lyrics = r.json()['lyrics']
            # If unavailable leave as empty string
            if lyrics_tp is None:
                lyrics = ""
        else:
            lyrics = ""
        return lyrics


import os
from mutagen.flac import FLAC, Picture
from mutagen.mp4 import MP4
from loguru import logger
from concurrent.futures import ThreadPoolExecutor

from rsack.clients.kkbox import KkboxAPI
from rsack.utils import Settings, sanitize

class Download():
    def __init__(self, url):
        self.client = KkboxAPI()
        self.settings = Settings().KKBox()
        self.id = url.split("/")[-1]
        self.client.login(email=self.settings['email'],
                            password=self.settings['password'],
                            region_bypass=True)
        if url.split("/")[-2] == "artist":
            artist = self.client.get_artist(self.id)
            for album in artist['album']:
                self._download_album(album['encrypted_album_id'])
        else:
            self._download_album(self.id)

    @staticmethod
    def _determine_type(audio_quality: str) -> str:
        """Return name of stream data given the audio quality
        """
        legend = {
            '128k': 'mp3_128k_chromecast',
            '192k': 'mp3_192k_kkdrm1',
            '320k': 'aac_320k_m4a_kkdrm1',
            'hifi': 'flac_16_download_kkdrm',
            'hires': 'flac_24_download_kkdrm',
        }
        return legend[audio_quality]
    
    def get_img_url(self, url_template, size, file_type='jpg'):
        url = url_template
        # not using .format() here because of possible data leak vulnerabilities
        if size > 2048:
            url = url.replace('fit/{width}x{height}', 'original')
            url = url.replace('cropresize/{width}x{height}', 'original')
        else:
            url = url.replace('{width}', str(size))
            url = url.replace('{height}', str(size))
        url = url.replace('{format}', file_type)
        return url
    
    @logger.catch
    def _template(self):
        keys = {
            "artist": self.meta['album']['artist_name'],
            "title": self.meta['album']['album_name'],
            "date": self.meta['release_date'],
            "album_id": str(self.meta['album']['album_id']),
            "encrypted_album_id": self.meta['album']['encrypted_album_id'],
            "artist_id": str(self.meta['album']['artist_id']),
            "encrypted_artist_id": self.meta['album']['encrypted_artist_id']
        }
        template = self.settings['template']
        for k in keys:
            template = template.replace(f"{{{k}}}", sanitize(keys[k]))
        return template
    
    def _create_album_folder(self):
        self.album_path = self.settings['path'] + self._template()
        try:
            if not os.path.exists(self.album_path):
                os.makedirs(self.album_path)
        except OSError as exc:
            if exc.errno == 36: # Exceeded path limit
                self.album_path = os.path.join(self.settings['path'], 'EDIT ME')
                if not os.path.exists(self.album_path): # Retry
                    os.makedirs(self.album_path)
    
    def _download_cover(self):
        url = self.get_img_url(self.meta['album']['album_photo_info']['url_template'], 3000)
        logger.info("Downloading original artwork")
        r = self.client.s.get(url)
        self.cover_path = os.path.join(self.album_path, 'cover.jpg')
        with open(self.cover_path, 'wb') as f:
            f.write(r.content)
            
        logger.info("Downloading artwork to embed")
        url = self.get_img_url(self.meta['album']['album_photo_info']['url_template'], 600)
        r = self.client.s.get(url)
        self.cover_path_embed = os.path.join(self.album_path, 'embed.jpg')
        with open(self.cover_path_embed, 'wb') as f:
            f.write(r.content)
    
    @logger.catch
    def _download_album(self, id):
        self.meta = self.client.get_album(id=id)
        self.meta['release_date'] = self.client.get_date(id)
        song_list = self.client.get_album_more(self.meta['album']['album_id'])
        self._create_album_folder()
        self._download_cover()
        self.meta['album']['track_total'] = song_list['song_list']['song'][-1]['trankno'] # Assign track total because the existing 'collected_count' is not accurate.
        with ThreadPoolExecutor(max_workers=int(self.settings['threads'])) as executor:
            executor.map(self._download_track, song_list['song_list']['song'])
        logger.debug(f"Deleting {self.cover_path_embed}")
        os.remove(self.cover_path_embed)
    
    def _format_lyrics(self, lyrics: dict):
        embedded = ''
        synced = ''
        for lyr in lyrics['data']['lyrics']:
            if not lyr['content']:
                embedded += '\n'
                synced += '\n'
                continue

            time = lyr['start_time']
            min = int(time / (1000 * 60))
            sec = int(time / 1000) % 60
            ms = int(time % 100)
            time_tag = f'[{min:02d}:{sec:02d}.{ms:02d}]'

            embedded += lyr['content'] + '\n'
            synced += time_tag + lyr['content'] + '\n'
        return synced
    
    @logger.catch
    def _download_track(self, song: dict):
        id = song['song_more_url'].split('/')[-1]
        urls = self.client.get_ticket(id, "webclient")
        url_type = self._determine_type(song['audio_quality'][-1])
        for u in urls:
            if u['name'] == url_type:
                url = u['url']
        file_ext = f".{url_type.split('_')[0]}"
        if file_ext == '.aac': # Correct .aac to .m4a
            file_ext = '.m4a'
        file_name = sanitize(f"{song['trankno'].zfill(2)}. {song['text']}{file_ext}")
        file_path = os.path.join(self.album_path, file_name)
        if not os.path.exists(file_path):
            logger.info(f"{file_name}")
            self.client.kkdrm_dl(url=url, path=file_path)
            if song['song_lyrics_valid'] == 1:
                l = self.client.get_song_lyrics(id)
                if l['status']['type'] ==  "OK":
                    lyrics = self._format_lyrics(l)
                else:
                    lyrics = ''
            else:
                lyrics = ''
            Tag(file_path=file_path, file_ext=file_ext, track_meta=song, lyrics=lyrics, album_meta=self.meta, cover_path=self.cover_path_embed)
        else:
            logger.info(f"Already exists: {file_path}")
            
class Tag():
    def __init__(self, file_path: str, file_ext: str, track_meta: dict, lyrics: str, album_meta: dict, cover_path: str):
        self.cover_path = cover_path
        self.file_path = file_path
        self.album_meta = album_meta
        self.track_meta = track_meta
        self.lyrics = lyrics
        if file_ext == ".flac":
            self.flac()
        elif file_ext == '.m4a':
            self.aac()
    
    @logger.catch
    def aac(self):
        audio = MP4(self.file_path)

        # Tags
        logger.debug(f"Writing tags to {self.file_path}")
        audio['\xa9ART'] = self.track_meta['artist_role']['mainartist_list']['mainartist']
        audio['aART'] = self.album_meta['album']['artist_name'] # Look into this
        audio['\xa9alb'] = self.album_meta['album']['album_name']
        audio['\xa9nam'] = self.track_meta['text']
        audio['\xa9cmt'] = self.track_meta['song_id']
        audio['\xa9gen'] = self.track_meta['genre_name']
        audio['trkn'] = [(int(self.track_meta['trankno']), int(self.album_meta['album']['track_total']))]
        audio['disk'] = [(1, 1)] # KKBOX don't include disc numbers
        audio['\xa9day'] = self.album_meta['release_date']
        
        audio.save()
        
    @logger.catch
    def flac(self):
        """Tag flac file
        """
        audio = FLAC(self.file_path) # Initialize FLAC object.
        
        # Remove replay gain tags
        audio.pop('replaygain_track_peak')
        audio.pop('replaygain_album_gain')
        audio.pop('replaygain_track_gain')
        audio.pop('replaygain_album_peak')
        audio.pop('replaygain_reference_loudness')

        # Embed cover artwork
        audio_image = Picture()
        audio_image.type = 3
        audio_image.desc = 'Front Cover'
        with open(self.cover_path, 'rb') as f:
            audio_image.data = f.read()
        audio.add_picture(audio_image)
        
        # Tags
        logger.debug(f"Writing tags to {self.file_path}")
        audio['ARTIST'] = self.track_meta['artist_role']['mainartist_list']['mainartist']
        audio['ALBUMARTIST'] = self.album_meta['album']['artist_name'] # Look into this
        audio['ALBUM'] = self.album_meta['album']['album_name']
        audio['TITLE'] = self.track_meta['text']
        audio['COMMENT'] = self.track_meta['song_id']
        audio['GENRE'] = self.track_meta['genre_name']
        audio['TRACKNUMBER'] = self.track_meta['trankno']
        audio['TRACKTOTAL'] = str(self.album_meta['album']['track_total'])
        audio['DISCTOTAL'] = "1" # KKBOX don't include disc totals
        audio['DISCNUMBER'] = "1" # KKBOX don't include disc numbers
        audio['DATE'] = self.album_meta['release_date']
        
        if self.lyrics != '':
            audio['LYRICS'] = self.lyrics
                
        # Save file
        audio.save()


from requests.models import Response

class DeviceIDError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self) -> str:
        return self.message

class InvokeMapError(Exception):
    def __init__(self, response: Response):
        self.message = response['ret_detail_msg']
    
    def __str__(self) -> str:
        return self.message
    
class InvalidURL(Exception):
    def __init__(self, url: str):
        self.url = url
    
    def __str__(self) -> str:
        return f"{self.url} is invalid."    


# Standard
import requests
from loguru import logger

from rsack.exceptions import InvokeMapError

class Client:
    def __init__(self):
        self.session = requests.Session()
        self.api_key = "b2de0fbe3380408bace96a5d1a76f800"
        self.session.headers.update({
            "User-Agent": "Mobile|Bugs|4.11.30|Android|5.1.1|SM-G965N|samsung|market",
            "Host": "api.bugs.co.kr",
        })
 
    def auth(self, email: str, password: str):
        """Authenticates session"""
        data = {
            "device_id": "gwAHWlkOYX_T8Sl43N78GiaD6Sg_", # Hardcode device id
            "passwd": password,
            "userid": email
        }
        r = self.make_call("secure", "mbugs/3/login?", data=data)
        if r['ret_code'] == 300:
            logger.critical("Authentication Error, Invalid Credentials")
        else:
            logger.info(f"Login Successful : {r['result']['right']['product']['name']}")
        self.nickname = r['result']['extra_data']['nickname']
        self.connection_info = r['result']['coninfo']
        self.premium = r['result']['right']['stream']['is_flac_premium']
        return self.connection_info

    def make_call(self, sub: str, epoint: str, data: dict = None, json: dict = None, params: dict = None):
        """Makes an API call

        Args:
            sub (str): Subdomain
            epoint (str): Endpoint
            data (dict, optional): POST data. Defaults to None.
            json (dict, optional): POST json. Defaults to None.
            params (dict, optional): POST parameters. Defaults to None.

        Returns:
            dict: Response
        """
        r = self.session.post("https://{}.bugs.co.kr/{}api_key={}".format(sub, epoint, self.api_key), json=json, data=data, params=params)
        return r.json()

    def get_artist(self, id: int) -> dict:
        """Retrieves artist information

        Args:
            id (int): Artists unique id

        Raises:
            InvokeMapError: Failed to invoke map

        Returns:
            dict: API response
        """
        json = [{
            "id": "artist_info",
            "args": {"artistId": id}
        },
                {
                    "id": "artist_album",
                    "args": {"artistId": id,
                             "albumType": "main",
                             "tracksYn": "Y",
                             "page": 1,
                             "size": 500
                            }}]
        r = self.make_call("api", "3/home/invokeMap?", json=json)
        if r['ret_code'] != 0:
            raise InvokeMapError(r)
        return r
    
    def get_album(self, id: int) -> dict:
        """Retrieves album information

        Args:
            id (int): Album unique id.

        Raises:
            InvokeMapError: Failed to invoke map.

        Returns:
            dict: API response.
        """
        json = [{
            "id": "album_info",
            "args": {"albumId": id}
        },
                {
                    "id": "artist_role_info",
                    "args": {"contentsId": id,
                             "type": "ALBUM"
                            }}]
        r = self.make_call("api", "3/home/invokeMap?", json=json)
        if r['ret_code'] != 0:
            raise InvokeMapError(r)
        return r
    
    def get_track(self, id: int) -> dict:
        """Retrieves track information

        Args:
            id (int): Track unique id.

        Raises:
            InvokeMapError: Failed to invoke map.

        Returns:
            dict: API reponse
        """
        json=[{"id":"track_detail", 
               "args":{"trackId":id}
            }]
        r = self.make_call("api", "3/home/invokeMap?", json=json)
        if r['ret_code'] != 0:
            raise InvokeMapError(r)
        return r


from setuptools import setup, find_packages
from rsack.version import __version__

setup(
    name="rsack",
    description="A Multi-purpose downloader.",
    long_description="Read more at https://github.com/Slyyxp/rsack",
    url="https://github.com/Slyyxp/rsack",
    author="Slyyxp",
    author_email="slyyxp@protonmail.com",
    version=__version__,
    packages=find_packages(),
    install_requires=["requests==2.25.1", "mutagen==1.45.1", "loguru==0.5.3"],
    entry_points={
                'console_scripts': [
                    'rsack = rsack.main:main'
                ]
    })



from loguru import logger
from argparse import ArgumentParser
from urllib.parse import urlparse

from rsack.version import __version__
from rsack import bugs, genie, kkbox
from rsack.utils import Settings, bugs_id, genie_id

def get_args():
    """Generate arguments"""
    parser = ArgumentParser()
    parser.add_argument('-v', '--version', action='store_true', dest='version', required=False)
    parser.add_argument('-u', '--url', nargs='*', dest="url", required=False)
    return parser.parse_args()

def main():
    """Entry point"""
    args = get_args()
    if args.version:
        print(__version__)
    if args.url:
        for url in args.url:
            domain = urlparse(url).netloc.replace("www.", "")
            if domain == "music.bugs.co.kr":
                id = bugs_id(url)
                if "album" in url:
                    bugs.Download(type="album", id=int(id))
                elif "artist" in url:
                    bugs.Download(type="artist", id=int(id))
                elif "track" in url:
                    bugs.Download(type="Single tracks not yet available", id=int(id))
            elif domain == "genie.co.kr":
                    match = genie_id(url)
                    if match.group(1) == "artistInfo":
                        type = "artist"
                    elif match.group(1) == "albumInfo":
                        type = "album"
                    else: # Catch invalid info types
                        logger.critical("URL type unkown")
                    genie.Download(type=type, id=int(match.group(2)))
            elif domain == "play.kkbox.com" or domain == "kkbox.com":
                kkbox.Download(url)